/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.performance.consumer;

import java.util.logging.Level;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.servicedesc.ServiceEndpoint;

/**
 * @author alouis - EBM WebSourcing
 * 
 */
public class AckTask implements Runnable {

    /**
     * 
     */
    protected Sender sender;

    protected ComponentContext context;
    protected DeliveryChannel channel;
    protected MessageExchangeFactory factory;
    protected ServiceEndpoint providerEndpoint;

    /**
     * Creates a new instance of SenderThread
     * 
     * @param monitoring
     * @param startSignal
     * @param doneSignal
     */
    public AckTask(Sender sender) {
        this.sender = sender;
        context = sender.context;
        channel = sender.channel;
        factory = channel.createExchangeFactory();
        providerEndpoint = sender.providerEndpoint;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Thread#run()
     */
    public void run() {
        //accept (not blocking) exchange (receive done status)
        // accepted exchange can be different from the sent one.
        accept();
    }
    
    
 
    private void accept() {
        try{
        MessageExchange receivedExchange = channel.accept(5000);
        
        if(receivedExchange != null) {
            if (receivedExchange.getProperty("tracked") != null &&
                    ((String)receivedExchange.getProperty("tracked")).equals("true") ) {
                sender.acceptMarker.write(receivedExchange.getExchangeId() + "\t" + System.currentTimeMillis()+"\n");
            }
            
            if(ExchangeStatus.DONE.equals(receivedExchange.getStatus())){
                sender.incDoneExchangeCount();
            } else if(ExchangeStatus.ERROR.equals(receivedExchange.getStatus())){
                sender.incErrorExchangeCount();
            } else {
                if(receivedExchange.getFault() != null){
                    sender.incErrorExchangeCount();
                } else{
                   receivedExchange.setStatus(ExchangeStatus.DONE);
                   channel.send(receivedExchange);
                }
            }
        }
        } catch (Throwable e) {
            sender.logger.log(Level.SEVERE,e.getMessage(),e);
        }
    }

}
