package org.objectweb.petals.engine.performance.consumer;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Collection;
import java.util.HashSet;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

public class Sender implements SenderMBean {

    protected static final String IN_ONLY = "IN_ONLY";
    protected static final String IN_OUT = "IN_OUT";
    
    protected String operation;

    protected boolean sync = false;
    
    protected int syncTimeout;

    protected String mep = IN_ONLY;
    
    protected boolean acknowledgment = true;

    /*
     * The content of the message can be either fixed, random, or imported from an
     * external file
     */ 
    protected int size = 1000;
    protected boolean randomSize = false;
    protected String contentFileName = null;
    
    protected boolean compression = false;

    protected boolean persistent = true;

    protected int threadCount = 4;

    protected String duration = "10000";

    protected long trackingFrequencyPerMinute = 6000;
    
    protected boolean running = false;
    
    private long sentMessageCount;
    private long notSentMessageCount;
    private long doneExchangeCount;
    private long errorExchangeCount;
    
    /**
     * Non mbean exposed
     */
    private String providerInterface = "{http://petals.objectweb.org}PerfProvider";

    protected Logger logger;
    protected ComponentContext context;
    protected DeliveryChannel channel;
    protected ServiceEndpoint providerEndpoint;
    
    protected ScheduledThreadPoolExecutor scheduledThreadPoolExecutor;
    protected ScheduledThreadPoolExecutor ackScheduledThreadPoolExecutor;
    
    protected BufferedWriter beforeSendMarker;
    protected BufferedWriter afterSendMarker;
    protected BufferedWriter acceptMarker;
    
    protected String data;
    protected long lastTrackedMessageTime;
    //laps time between each tracked object
    protected long trackingLapsTime = 100;
    

    public Sender() {
        setSize(size);
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setAcknowledgment(boolean)
     */
    public void setAcknowledgment(boolean acknowledgment) {
        this.acknowledgment=acknowledgment;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#isAcknowledgment()
     */
    public boolean isAcknowledgment() {
        return acknowledgment;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#isRunning()
     */
    public boolean isRunning() {
        return running;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getMep()
     */
    public String getMep() {
        return mep;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setMep(java.lang.String)
     */
    public void setMep(String mep) {
        this.mep = mep;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getOperation()
     */
    public String getOperation() {
        return operation;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setOperation(java.lang.String)
     */
    public void setOperation(String operation) {
        this.operation = operation;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getSize()
     */
    public int getSize() {
        return size;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setSize(int)
     */
    public void setSize(int size) {
        this.size = size;
        StringBuffer result = new StringBuffer(size);
        result.append("<data>");
        for (int i = 0; i < size - 13; i++) {
            result.append("a");
        }
        result.append("</data>");
        data = result.toString();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#isSync()
     */
    public boolean isSync() {
        return sync;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setSync(boolean)
     */
    public void setSync(boolean sync) {
        this.sync = sync;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getThreadCount()
     */
    public int getThreadCount() {
        return threadCount;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setThreadCount(int)
     */
    public void setThreadCount(int threadCount) {
        this.threadCount = threadCount;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getSyncTimeout()
     */
    public int getSyncTimeout() {
        return syncTimeout;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setSyncTimeout(int)
     */
    public void setSyncTimeout(int syncTimeout) {
        this.syncTimeout = syncTimeout;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getProviderInterface()
     */
    public String getProviderInterface() {
        return providerInterface;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setProviderInterface(java.lang.String)
     */
    public void setProviderInterface(String providerInterface) {
        this.providerInterface = providerInterface;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getDuration()
     */
    public String getDuration() {
        return duration;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setDuration(java.lang.String)
     */
    public void setDuration(String duration) {
        this.duration = duration;
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#isPersistent()
     */
    public boolean isPersistent() {
        return persistent;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setPersistent(boolean)
     */
    public void setPersistent(boolean persistent) {
        this.persistent = persistent;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getSentMessageCount()
     */
    public long getSentMessageCount() {
        return sentMessageCount;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getSentNotMessageCount()
     */
    public long getNotSentMessageCount() {
        return notSentMessageCount;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getDoneExchangeCount()
     */
    public long getDoneExchangeCount() {
        return doneExchangeCount;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getErrorExchangeCount()
     */
    public long getErrorExchangeCount() {
        return errorExchangeCount;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#isRandomSize()
     */
    public boolean isRandomSize() {
        return randomSize;
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setRandomSize(boolean)
     */
    public void setRandomSize(boolean randomSize) {
        this.randomSize = randomSize;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getTrackingFrequencyPerMinute()
     */
    public long getTrackingFrequencyPerMinute() {
        return trackingFrequencyPerMinute;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setTrackingFrequencyPerMinute(long)
     */
    public void setTrackingFrequencyPerMinute(long trackingFrequencyPerMinute) {
        this.trackingFrequencyPerMinute = trackingFrequencyPerMinute;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#isCompression()
     */
    public boolean isCompression() {
        return compression;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#setCompression(boolean)
     */
    public void setCompression(boolean compression) {
        this.compression = compression;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.performance.component.consumer.SenderMBean#getContentFileName()
     */
    public String getContentFileName() {
        return contentFileName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.engine.performance.consumer.SenderMBean#setContentFileName(java.lang.String)
     */
    public void setContentFileName(String contentFileName) {
        this.contentFileName = contentFileName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.engine.performance.consumer.SenderMBean#startSending()
     */
    public void startSending() throws Exception {

        try {
            // get the logs directory
            File petalsroot = new File(context.getInstallRoot())
                .getParentFile().getParentFile().getParentFile()
                .getParentFile();
            File logs = new File(petalsroot, "logs");

            // init log files
            beforeSendMarker = new BufferedWriter(new FileWriter(new File(logs,
                "consumer-beforeSend.csv")));
            afterSendMarker = new BufferedWriter(new FileWriter(new File(logs,
                "consumer-afterSend.csv")));
            acceptMarker = new BufferedWriter(new FileWriter(new File(logs,
                "consumer-accept.csv")));

            // find the endpoint
            ServiceEndpoint[] providers = context.getEndpoints(QName
                .valueOf(providerInterface));
            if (providers != null && providers.length > 0) {
                providerEndpoint = providers[0];
            } else {
                String msg = "No Endpoint satisfies the interface: "
                    + providerInterface;
                System.out.println(msg);
                throw new Exception(msg);
            }

            if ((!IN_ONLY.equalsIgnoreCase(mep))
                && (!IN_OUT.equalsIgnoreCase(mep))) {
                String msg = "MEP must be IN_ONLY or IN_OUT.";
                throw new Exception(msg);
            }
            // create sending-threads
            Collection<SendTask> tasks = new HashSet<SendTask>();
            int i = 0;
            while (i++ < threadCount) {
                tasks.add(new SendTask(this));
            }

            trackingLapsTime = 60000 / trackingFrequencyPerMinute;

            if (contentFileName != null) {
                File contentFile = new File(contentFileName);
                if (!contentFile.exists()) {
                    throw new Exception("ContentFile doesn't exist");
                }
                BufferedReader inputReader = new BufferedReader(
                    new InputStreamReader(new FileInputStream(contentFile),
                        "UTF-8"), 1024);
                char[] bufferChars = new char[1024];
                StringBuilder stringBuilder = new StringBuilder(1024);
                int charsLengh;
                while ((charsLengh = inputReader.read(bufferChars)) != -1) {
                    String bufferString = String.valueOf(bufferChars, 0,
                        charsLengh);
                    stringBuilder.append(bufferString);
                }
                data = stringBuilder.toString();
            }

            // let the threads works until "duration" elapsed
            final long dure = Long.parseLong(duration);

            scheduledThreadPoolExecutor = new ScheduledThreadPoolExecutor(
                threadCount);

            //set the current date before starting tracking the messages 
            lastTrackedMessageTime = System.currentTimeMillis();

            for (SendTask task : tasks) {
                scheduledThreadPoolExecutor.scheduleWithFixedDelay(task, 0, 1,
                    TimeUnit.MILLISECONDS);
            }

            // start an Acknowledgment thread
            AckTask ackTask = new AckTask(this);
            ackScheduledThreadPoolExecutor = new ScheduledThreadPoolExecutor(1);
            ackScheduledThreadPoolExecutor.scheduleWithFixedDelay(ackTask, 0,
                1, TimeUnit.MILLISECONDS);

            Runnable r = new Runnable() {
                public void run() {
                    try {
                        Thread.sleep(dure);
                    } catch (InterruptedException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                    try {
                        scheduledThreadPoolExecutor.shutdown();
                        scheduledThreadPoolExecutor.awaitTermination(10, TimeUnit.SECONDS);
                    } catch (InterruptedException e1) {
                        // do nothing
                    }
                    running = false;
                }
            };
            new Thread(r).start();
            running = true;
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            // TODO Auto-generated catch block
            throw e;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.engine.performance.consumer.SenderMBean#stopSending()
     */
    public void stopSending() {
        try {
            if ( !scheduledThreadPoolExecutor.isShutdown() ) {
                scheduledThreadPoolExecutor.shutdown();
                scheduledThreadPoolExecutor.awaitTermination(10, TimeUnit.SECONDS);
            }
            ackScheduledThreadPoolExecutor.shutdown();
            ackScheduledThreadPoolExecutor.awaitTermination(10, TimeUnit.SECONDS);
        } catch (InterruptedException e1) {
            // do nothing
        }
        try {
            beforeSendMarker.close();
            afterSendMarker.close();
            acceptMarker.close();
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

   /*
    * (non-Javadoc)
    * 
    * @see org.objectweb.petals.engine.performance.consumer.SenderMBean#reset()
    */
    public void resetCounters() {
        sentMessageCount = 0;
        notSentMessageCount = 0;
        errorExchangeCount = 0;
        doneExchangeCount = 0;
    }

    protected synchronized void incSentMessageCount() {
        sentMessageCount++;
    }

    protected synchronized void incNotSentMessageCount() {
        notSentMessageCount++;
    }

    protected synchronized void incDoneExchangeCount() {
        doneExchangeCount++;
    }

    protected synchronized void incErrorExchangeCount() {
        errorExchangeCount++;
    }
}
